@@ -12,12 +12,11 @@ import os
 from pathlib import Path
 from deprecated import deprecated
 from datetime import datetime, timedelta
+from numbers import Number
 from typing import (
     Any,
     TypeVar,
     Callable,
-    Union,
-    Optional,
 )
 
 
@@ -30,11 +29,11 @@ _F = TypeVar("_F", bound=Callable[..., Any])
 
 def public(
     *,
-    alias_for: Union[_F, None] = None,
+    alias_for: _T | None = None,
     class_alias: bool = False,
-    deprecate_reason: Optional[str] = None,
-    deprecate_dropat: Optional[str] = None,
-) -> Callable[[_F], _F]:
+    deprecate_reason: str | None = None,
+    deprecate_dropat: str | None = None,
+) -> Callable[[_T], _T]:
     """Documentation decorator, used to mark function/class which should be considered as a public API.
 
     Only elements marked with this decorator may be treated as `stable API`.
@@ -98,7 +97,7 @@ def public(
     deprecate_reason="Use `str(x.absolute()) inliner instead",
     deprecate_dropat="2.0",
 )
-def resolve(path: Path):
+def resolve(path: Path) -> str:
     return str(path.absolute())
 
 
@@ -106,7 +105,7 @@ def resolve(path: Path):
     deprecate_reason="Use `datetime.now().strftime('%Y-%m-%d %H:00:00')` instead.",
     deprecate_dropat="2.0",
 )
-def now(template: str = "%Y-%m-%d %H:00:00"):
+def now(template: str = "%Y-%m-%d %H:00:00") -> datetime:
     return datetime.now().strftime(template)
 
 
@@ -120,50 +119,82 @@ class _StreamOutputDumper(threading.Thread):
         callback: typing.Callable[[str], None],
     ):
         threading.Thread.__init__(self)
+        self.daemon = True
         self.process = process
         self.stream = stream
         self.callback = callback
+        self._last_line_incomplete = False
         self._result_list = []
+        self.done = False
         self.start()
 
+    def print_line(self, line: bytes):
+        self.callback(line.decode(errors='ignore'))
+
+    def print_line_incomplete(self, line: bytes):
+        if line:
+            self.callback(
+                line.decode(errors='ignore'),
+                extra={'incomplete_line': True},
+            )
+
     def result(self):
         self.join()
-        return "".join(self._result_list)
+        return b"".join(self._result_list).decode()
 
     def run(self):
-        space_buffer = []
 
-        while not self.stream.closed:
+        buffer = b""
+        delay = 0.001
+        incomplete_line = False
+
+        while True:
             try:
-                line = self.stream.readline()
+                raw: bytes = self.stream.read()
             except (ValueError, EOFError):
-                return  # closed
+                break  # closed
 
-            linee = line
-            if linee.endswith("\n"):
-                linee = linee[:-1]
+            if not raw:
+                if self.done:
+                    break
 
-            if linee.strip():
-                # log line and prepend all buffered whitespaces
-                space_buffer.append(linee)
-                self.callback("".join(space_buffer))
-                space_buffer.clear()
-            else:
-                # line contains only whitespaces
-                space_buffer.append(line)
-            self._result_list.append(line)
+                delay = min(1, delay * 1.66)
+                time.sleep(delay)
+
+                if delay > 0.01 and buffer and not incomplete_line:
+                    self.print_line_incomplete(buffer)
+                    incomplete_line = True
+
+                continue
+
+            delay = 0.001
+            self._result_list.append(raw)
+            buffer += raw
+
+            lines = buffer.split(b"\n")
+
+            for line in lines[:-1]:
+                incomplete_line = False
+                self.print_line(line)
+
+            buffer = lines[-1]
+            if buffer and incomplete_line:
+                self.print_line_incomplete(buffer)
+
+        for line in buffer.split(b"\n"):
+            self.print_line(line)
 
 
 def run_process(
-    args: typing.Union[str, typing.List],
+    args: str | list[str],
     *,
-    verbose=True,
-    check=True,
-    env_add=None,
-    input=None,
-    env=None,
+    verbose: bool = True,
+    check: bool = True,
+    env_add: dict[str, str] = None,
+    input: str | None = None,
+    env: dict[str, str] = None,
     **kwargs,
-):
+) -> str:
     """Run external process, extends `subprocess.run`, returns 'stdout', may throw `subprocess.SubprocessError`.
 
     Arguments:
@@ -183,14 +214,15 @@ def run_process(
             for x in args
         ]
 
-    logger.debug("cmd %r, kwargs %r", cmd, kwargs)
+    logger.debug("run process %r", cmd)
     if env_add:
         env = dict(env or os.environ)
         env.update(env_add)
 
     start = time.time()
     process = subprocess.Popen(
-        cmd, text=True,
+        cmd,
+        text=False,
         stdout=subprocess.PIPE,
         stderr=subprocess.PIPE,
         stdin=subprocess.PIPE if input is not None else None,
@@ -201,20 +233,31 @@ def run_process(
     assert process.stdout, "invalid stdout"
     assert process.stderr, "invalid stderr"
 
+    os.set_blocking(process.stdout.fileno(), False)
+    os.set_blocking(process.stderr.fileno(), False)
+
     stdout_dumper = _StreamOutputDumper(
         process, process.stdout, logger.info if verbose else logger.debug)
+
     stderr_dumper = _StreamOutputDumper(
         process, process.stderr, logger.error if verbose else logger.debug)
 
     if input:
         assert process.stdin, "invalid stdin"
+        if isinstance(input, str):
+            input = input.encode()
         process.stdin.write(input)
         process.stdin.close()
 
     code = process.wait()
+    stdout_dumper.done = True
+    stderr_dumper.done = True
+    stdout_dumper.join()
+    stderr_dumper.join()
 
     process.stdout.close()
     process.stderr.close()
+
     stdout = stdout_dumper.result()
     stderr = stderr_dumper.result()
 
@@ -249,9 +292,9 @@ def decode_version_number_from_file_name(file_path: Path):
 
 def get_docker_image_id(tag):
     logger.info("Getting docker image ID.")
-    images = run_process(f"docker images -q {tag}")
+    images = run_process(["docker", "images", "-q", "--no-trunc", tag])
     logger.info(images[:1000] + '...')
-    return images.split('\n')[0]
+    return images.split("\n")[0]
 
 
 def build_docker_image_tag(docker_repository: str, package_version: str):
@@ -260,18 +303,41 @@ def build_docker_image_tag(docker_repository: str, package_version: str):
 
 
 def remove_docker_image_from_local_registry(tag):
-    print('Removing the image from the local registry')
-    run_process(f"docker rmi -f {get_docker_image_id(tag)} --no-prune")
+    logger.info("Removing the image from the local registry")
+    image = get_docker_image_id(tag)
+    run_process(f"docker rmi -f {image} --no-prune")
+    logger.debug("Image %s removed from docker registry", image)
 
 
-def as_timedelta(v: typing.Union[None, str, int, float, timedelta]) -> typing.Optional[timedelta]:
+def as_timedelta(v: None | str | Number | timedelta) -> timedelta | None:
     if v is None:
         return None
     elif isinstance(v, timedelta):
         return v
-    elif isinstance(v, (int, float)):
-        return timedelta(seconds=v)
+    elif isinstance(v, Number):
+        return timedelta(seconds=float(v))
     elif v == "":
         return None
     else:
         return timedelta(seconds=float(v))
+
+
+def valid_datetime(dt: str) -> str:
+    """
+    Validates provided datetime string. Raises ValueError for strings that are none of:
+    * 'NOW'
+    * valid '%Y-%m-%d %H:%M:%S'
+    * valid '%Y-%m-%d'
+    """
+    if dt == 'NOW':
+        return dt
+
+    try:
+        datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")
+    except ValueError:
+        try:
+            datetime.strptime(dt, "%Y-%m-%d")
+        except ValueError:
+            raise ValueError("Not a valid date: '{0}'.".format(dt))
+
+    return dt
